/*
 * JonoF's Multiplayer for the Build Engine
 * Copyright (c) 2003 Jonathon Fowler
 * See BUILDLIC.TXT for license details.
 */

#define PEDANTIC
//#define FAKESEND
#define DUMPEACHPACKET
#define CHATTY
//#define ENABLECOMMDUMPING

// Communications dump blocks:
// Each block has an id byte and a 32-bit int representing the clock value
//  0 - Initialization
//  1 - Add player
//      char strlen
//      char playerhost[strlen]
//  2 - sendto() data
//      char multiaddrsz
//      multiaddr multiaddr
//      short datalen
//      char data[datalen]
//  3 - recvfrom() data
//      char multiaddrsz
//      multiaddr multiaddr
//      short datalen
//      char data[datalen]
//  4 - File closed


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "compat.h"
#include "osd.h"
#include "crc32.h"
#include "baselayer.h"

#ifdef WINDOWS
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
//#include <winsock2.h>
#include <winsock.h>
#define ER(x) WSA##x
#define ERS(x) "WSA" x
#define SHUT_RD 0
#define SHUT_WR 1
#define SHUT_RDWR 2
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#define ER(x) x
#define ERS(x) x
#endif

extern long totalclock;
static long lastclock;

static struct {
	unsigned long txbytes;
	unsigned long rxbytes;
	unsigned long txpackets;
	unsigned long rxpackets;
	unsigned long txerrors;
	unsigned long rxerrors;
} stats = { 0,0,0,0,0,0 };

// address structure; opaque to the game
#define ADDR_NONE (0)
#define ADDR_IPV4 (1)
#define ADDR_IPV6 (2)
typedef struct {
	unsigned short type;
	union {
		struct sockaddr_in ipv4;
		//struct sockaddr_in6 ipv6;
	} addr;
} multiaddr;


#define IN_MMULTI	// include this here so we define multiaddr beforehand
#include "mmulti.h"

/*
 Packet Structure:

 unsigned long  crc;			// crc32 of the entire packet data (including header, but excluding this property)
 unsigned char  messagecount;		// number of messages contained in this packet
 {
 unsigned char  sequence;		// sequence number of the message
 unsigned char  length;			// the length of the message minus 1
 unsigned char  flags;			// bits 0 & 1 represent message type
 unsigned char  messagedata[length+1];	// data for this message
 } * messagecount;			// sorted ascending by sequence number
 unsigned char  acklen;			// byte-length of unacknowledged list
 unsigned char  ack[];			// acknowledged sequence numbers, array size being calculated
					//   from (packetlength - headersize - messages - 1(acklen))
*/


// constants
#define MAX_MESSAGE_LEN	256
#define BACKLOGSZ	16384
#define NUM_SEQUENCES	256	// range 0-255

#define MESSAGEHEADLEN	(1+1+1)
#define PACKETHEADLEN	(4+1)

// completely arbitrary.
// best thing would be to actually determine the path MTU using path MTU discovery
// as described in 'man 7 ip'. Perhaps later
#define MAX_PACKET_LEN	((MAX_MESSAGE_LEN + MESSAGEHEADLEN) * 4 + PACKETHEADLEN + 16)

#define ALLOC_IN	0
#define ALLOC_OUT	1

#define FLAG_TYPE_SYNC   0
#define FLAG_TYPE_UNSYNC 1
#define FLAG_TYPE_OOB    2

#define comparesequences(a,op,b) (((signed short)(a) - (signed short)(b)) op 0)

// out-of-band input queue
#define OOBQUEUESZ	16
static struct {
	char          *ptr;	// pointer into inmsgbuf for 'len' many bytes
	unsigned short len;
	multiaddr      addr;
} oobmsgqueue[OOBQUEUESZ];
static unsigned char oobmsgqueuehead = 0, oobmsgqueuetail = 0;


// message queues
static char           outmsgbuf[BACKLOGSZ];
static unsigned long  outmsghead = 0;
static char           inmsgbuf[BACKLOGSZ];
static unsigned long  inmsghead = 0;


// peer structures
static struct {
	// outgoing messages
	//   outptr is a pointer into outmsgbuf, for outlen many bytes
	//   outlen == 0 means no message at the index
	char          *outptr[NUM_SEQUENCES];
	unsigned short outlen[NUM_SEQUENCES];
	unsigned char  outage[NUM_SEQUENCES];

	// last acknowledged sequence for messages sent to this peer
	unsigned char  acksequence;

	// next available sequence number for messages sent to this peer
	unsigned char  sequence;

	// outgoing oob messages
	//   oobptr is a pointer into outmsgbuf, for ooblen many bytes
	char          *oobptr[OOBQUEUESZ];
	unsigned short ooblen[OOBQUEUESZ];

	// head and tail for the oob message ring
	unsigned char  oobqueuehead, oobqueuetail;

	// incoming messages
	//   inptr is a pointer into inmsgbuf, for inlen many bytes
	//   inlen == 0 means no message at the index
	char          *inptr[NUM_SEQUENCES];
	unsigned short inlen[NUM_SEQUENCES];

	// next incoming message sequence
	unsigned char  insequence;

	// status flags for each sequence number
	//   bit 0 = marked for transmission in next packet
	//   bit 1 = requires acknowledgement
	unsigned char  msgflags[NUM_SEQUENCES];

	// the network address of this peer
	multiaddr      address;

	// peer flags
	unsigned char  flags;

	// network tuning parameters
	//   these parameters will one day allow the network code to
	//   vary its behaviour to compensate for network conditions
	//   like lag, loss, and bandwidth
} peers[MAXMULTIPLAYERS];

#define MSGFLAG_TX		1
#define MSGFLAG_ACK		2

#define PEERFLAG_FORCETX	1

// multiplayer operation state
char multioption = 0;	// single-player by default

// the connection number of the console
short myconnectindex;

// the number of players in the game
short numplayers;

// the head of the connection list
short connecthead;

// the next pointers of the connection list
short connectpoint2[MAXMULTIPLAYERS];

// uhmm...
char syncstate = 0;

// IP port number to communicate via
int ipport = 0x4A46;

static int ipsocket = -1;

#ifdef WINDOWS
// flags whether winsock has been loaded
static char winsockinited = 0;
#endif

static short fetchhead = 0;

#ifdef ENABLECOMMDUMPING
static BFILE *dumpfh = NULL;
static void putdumpblock(char c) {
	Bfputc(c, dumpfh);
	Bfputc(totalclock & 255, dumpfh);
	Bfputc((totalclock >> 8) & 255, dumpfh);
	Bfputc((totalclock >> 16) & 255, dumpfh);
	Bfputc((totalclock >> 24) & 255, dumpfh);
}
#endif

static int startnetwork(void);
static void stopnetwork(void);
static void networksend(multiaddr *addr, void *buf, unsigned long len);
static int networkrecv(multiaddr *addr, void *buf, unsigned long *len);

static int setupmultigame(void);
static void endmultigame(void);


/*
 * Internal trickery
 */

//
// resetpeer() -- Zeroes out a peer
//
static void resetpeer(int p)
{
	int i;

	for (i=0;i<NUM_SEQUENCES;i++) {
		peers[p].outptr[i] = 0;
		peers[p].outlen[i] = 0;
		peers[p].outage[i] = 0;
			
		peers[p].inptr[i] = 0;
		peers[p].inlen[i] = 0;

		peers[p].msgflags[i] = 0;
	}

	peers[p].acksequence = 0;
	peers[p].sequence = 0;

	peers[p].oobqueuehead = 0;
	peers[p].oobqueuetail = 0;
	
	peers[p].insequence = 0;

	peers[p].address.type = ADDR_NONE;

	peers[p].flags = 0;

#ifdef CHATTY
	printf("Zeroed peer %d\n", p);
#endif
}

//
// allocatestorage() -- Allocates storage on one of the ring buffers for packets
//
static unsigned char * allocatestorage(int ring, short leng)
{
	int ptr;
	
	switch (ring) {
		case ALLOC_IN:
			ptr = inmsghead;
			inmsghead = (inmsghead + leng) % BACKLOGSZ;

#ifdef CHATTY
			printf("allocatestorage(): Allocated %d inbytes, head now at %d\n", leng, inmsghead);
#endif
			
			return (unsigned char *)(inmsgbuf + ptr);

		case ALLOC_OUT:
			ptr = outmsghead;
			outmsghead = (outmsghead + leng) % BACKLOGSZ;

#ifdef CHATTY
			printf("allocatestorage(): Allocated %d outbytes, head now at %d\n", leng, outmsghead);
#endif
			
			return (unsigned char *)(outmsgbuf + ptr);

		default: return NULL;
	}
}

//
// agemessages() -- Ages any messages currently in peer outgoing backlogs
//
static void agemessages(void)
{
	int h,delta=0,j;
	unsigned char i;
	
	delta = totalclock - lastclock;
	lastclock = totalclock;
	if (delta <= 0) return;

#ifdef CHATTY
	printf("agemessages(): Time moved %d tics (now %d)\n", delta, totalclock);
#endif
	
	for (h=connecthead; h!=-1; h=connectpoint2[h]) {
		for (i=peers[h].acksequence;
		     i!=peers[h].sequence;
		     /*i=(i+1)%NUM_SEQUENCES*/ i++) {
			if (!peers[h].outlen[i]) continue;
			j = peers[h].outage[i] + delta;
			peers[h].outage[i] = j>255?255:j;

			// if message has aged past the retransmission threshold,
			// flag for retransmission
			if (j > 2) peers[h].msgflags[i] |= MSGFLAG_TX;
		}
	}
}

//
// processincoming() -- Fetches packets from the network and processes them
//
#define pktcrc(p) \
	( ((unsigned long)p[0]) | ((unsigned long)p[1]<<8) | ((unsigned long)p[2]<<16) | ((unsigned long)p[3]<<24) )
#define pktmsg(p) \
	( p[4] )
static void processincoming(void)
{
	int rv;
	multiaddr addr;
	short peer;
	unsigned char count;
	
	unsigned char workbuf[MAX_PACKET_LEN], *workptr, *p;
	unsigned long workbuflen;

	unsigned long crc;

	unsigned char sequence;
	unsigned short length;
	unsigned char flags;

	unsigned char unacklistlen, acklistlen;

	int i,j;

	// fetch a packet
	workbuflen = MAX_PACKET_LEN;
	if ((rv = networkrecv(&addr, workbuf, &workbuflen)) > 0) {
#ifdef CHATTY
		printf("processincoming(): Received %d bytes\n", workbuflen);
#endif
		if (workbuflen <= PACKETHEADLEN) {
#ifdef CHATTY
			printf("processincoming(): Packet was incomplete\n");
#endif
			stats.rxerrors++;
			return;
		}
		
		// check the CRC
		crc = crc32(workbuf+4, workbuflen-4);
		if (crc != pktcrc(workbuf)) {
#ifdef CHATTY
			printf("processincoming(): Packet was corrupt\n");
#endif
			stats.rxerrors++;
			return;		// corrupt or truncated
		}

		// determine which, if any, peer this came from
		peer = whichpeer(&addr);
#ifdef CHATTY
		if (peer < 0) {
			char temp[64];
			multiaddrtostring(&addr, temp, 64);
			printf("processincoming(): Host %s\n", temp);
		} else
			printf("processincoming(): Peer %d\n", peer);
#endif

		// add any messages to the appropriate queues
		workptr = workbuf + PACKETHEADLEN;
		for (count = 0; count < pktmsg(workbuf); count++, workptr += (MESSAGEHEADLEN + length)) {
			sequence = workptr[0];
			length = (unsigned short)workptr[1] + 1;
			flags = workptr[2];

			if ((workptr - workbuf) + (MESSAGEHEADLEN + length) > workbuflen) {
				// error in packet. length blows out the true length of the packet
#ifdef CHATTY
				printf("processincoming(): Packet is malformed.\n");
#endif
				stats.rxerrors++;
				break;
			}

			if ((flags & 3) == FLAG_TYPE_SYNC) {
				if (peer < 0) {
					// can't take game messages from a host not in the game
#ifdef CHATTY
					printf("processincoming(): Dropping alien synchronous message %d.\n", sequence);
#endif
					continue;
				}
				
				if (peers[peer].inlen[ sequence ] > 0) {
					// message already pending for this sequence. drop it on the floor and remind peer
					// we don't need another copy
					peers[peer].msgflags[sequence] |= MSGFLAG_ACK;	// flag for acknowledgement
#ifdef CHATTY
					printf("processincoming(): Duplicate message %d.\n", sequence);
#endif
				} else if (comparesequences(sequence, <, peers[peer].insequence)) {
					// message already consumed. remind the peer we don't need it anymore
					peers[peer].msgflags[sequence] |= MSGFLAG_ACK;	// flag for acknowledgement
#ifdef CHATTY
					printf("processincoming(): Redundant message %d.\n", sequence);
#endif
				} else {
					// queue message
					p = allocatestorage(ALLOC_IN, length);
					if (!p) continue;

					memcpy(p, workptr+MESSAGEHEADLEN, length);

					peers[peer].inptr[sequence] = p;
					peers[peer].inlen[sequence] = length;
					
					peers[peer].msgflags[sequence] |= MSGFLAG_ACK;	// flag for acknowledgement
					peers[peer].flags |= PEERFLAG_FORCETX;		// make sure the ack goes out

#ifdef CHATTY
					printf("processincoming(): Accepted message %d into queue.\n", sequence);
#endif
				}
			} else if ((flags & 3) == FLAG_TYPE_OOB) {
				if ((oobmsgqueuetail + 1) % OOBQUEUESZ == oobmsgqueuehead) {
#ifdef CHATTY
					printf("processincoming(): OOB ingress queue full.\n");
#endif
					continue;	// full
				}

				p = allocatestorage(ALLOC_IN, length);
				if (!p) continue;

				memcpy(p, workptr+MESSAGEHEADLEN, length);
				
				oobmsgqueue[oobmsgqueuetail].ptr = p;
				oobmsgqueue[oobmsgqueuetail].len = length;
				memcpy(&oobmsgqueue[oobmsgqueuetail].addr, &addr, sizeof(multiaddr));

				oobmsgqueuetail = (oobmsgqueuetail + 1) % OOBQUEUESZ;

#ifdef CHATTY
				printf("processincoming(): Accepted OOB message into queue.\n");
#endif
			}
		}

		// process acknowledgement lists
		if (peer != -1 && (workptr - workbuf) < workbuflen) {
			unacklistlen = *(workptr++);
			acklistlen = (workbuflen - (workptr - workbuf)) - unacklistlen;
#ifdef CHATTY
			printf("processincoming(): Unack list length %db, ack list length %db.\n",unacklistlen,acklistlen);
#endif

			if (unacklistlen > 0) {
				// If:
				//  j < insequence, it's already acknowledged, so just acknowledge again
				//  insequence <= j and incomingqueue[j] > 0, acknowledge
				j = *(workptr++); unacklistlen--;
				if ( comparesequences( j, <, peers[peer].insequence ) ||
				     ( comparesequences( peers[peer].insequence, <=, j ) &&
				       peers[peer].inlen[j] > 0 )
				   ) {
#ifdef CHATTY
					printf("processincoming(): Acking msg %d\n", j);
#endif
					peers[peer].msgflags[j] |= MSGFLAG_ACK;
					peers[peer].flags |= PEERFLAG_FORCETX;	// force ack transmission
				}
				for (; unacklistlen > 0; unacklistlen--,workptr++) {
					for (i=1;i<0x100;i<<=1) {
						j++;
						j&=255;
						if ((*workptr) & i) {
							if ( comparesequences( j, <, peers[peer].insequence ) ||
							     ( comparesequences( peers[peer].insequence, <=, j ) &&
							       peers[peer].inlen[j] > 0 )
							   ) {
#ifdef CHATTY
								printf("processincoming(): Acking msg %d\n", j);
#endif
								peers[peer].msgflags[j] |= MSGFLAG_ACK;
							}
						}
					}
				}
			}
			if (acklistlen > 0) {
				j = *(workptr++); acklistlen--;
				if ( comparesequences( peers[peer].acksequence, <=, j) ) {
#ifdef CHATTY
					printf("processincoming(): Got ack for msg %d\n", j);
#endif
					peers[peer].outlen[j] = 0;
					if (j == peers[peer].acksequence)
						while (peers[peer].outlen[ peers[peer].acksequence ] == 0 &&
						       peers[peer].acksequence != peers[peer].sequence )
							peers[peer].acksequence++;
				}
				for (; acklistlen > 0; acklistlen--,workptr++) {
					for (i=1;i<0x100;i<<=1) {
						j++;
						j&=255;
						if ((*workptr) & i) {
							if ( comparesequences( j, <, peers[peer].acksequence) ) continue;	// late ack
							
#ifdef CHATTY
							printf("processincoming(): Got ack for msg %d\n", j);
#endif
							peers[peer].outlen[j] = 0;
							if (j == peers[peer].acksequence)
								while (peers[peer].outlen[ peers[peer].acksequence ] == 0 &&
						                       peers[peer].acksequence != peers[peer].sequence)
									peers[peer].acksequence++;
						}
					}
				}
			}
		}
	}
}

//
// processoutgoing() -- Assembles packets and sends them
//
static void processoutgoing(void)
{
	int h,i,j,k,l;
	unsigned char s;

	unsigned char workbuf[MAX_PACKET_LEN], *workbufptr;
	unsigned char msgcount;

	unsigned char todo=0xff,resume=0;
	unsigned long crc;

	unsigned long myclock;

	memset(workbuf, 0, MAX_PACKET_LEN);
	
	// age any current messages and flag those which are stale
	myclock = lastclock;
	agemessages();

	// for each peer
	for (h=connecthead; h!=-1; h=connectpoint2[h]) {
		resume = 0;
		todo = 0xff;

resumepoint:
		workbufptr = workbuf + PACKETHEADLEN;
		msgcount = 0;
		
		if (todo & 1) {
			// if any messages flagged for transmission add them to the packet
			if (!resume) s=peers[h].acksequence;
			for (; s!=peers[h].sequence; s++) {
				if (peers[h].outlen[s] == 0) continue;
				if ((peers[h].msgflags[s] & MSGFLAG_TX) == 0) continue;
				if ((workbufptr - workbuf) + peers[h].outlen[s] > MAX_PACKET_LEN) {
					// data won't fit, so send the current packet on its way
#ifdef CHATTY
					printf("processoutgoing(): Packet full adding messages.\n");
#endif
					resume = 1;
					goto dispatch;
				} else {
					// pack the data in
					memcpy(workbufptr, peers[h].outptr[s], peers[h].outlen[s]);
					workbufptr += peers[h].outlen[s];
					msgcount++;
					peers[h].outage[s] = 0;
					peers[h].msgflags[s] &= ~MSGFLAG_TX;
#ifdef CHATTY
					printf("processoutgoing(): Added %d seqbytes to packet (msg %d)\n",peers[h].outlen[s],s);
#endif
				}
			}
			resume = 0;
			todo &= ~1;
		}

		if (todo & 2) {
			// if any oob messages queued add them to the packet
			while (peers[h].oobqueuehead != peers[h].oobqueuetail) {
				if ((workbufptr - workbuf) + peers[h].ooblen[ peers[h].oobqueuehead ] > MAX_PACKET_LEN) {
					// message won't fit, so send the current packet on its way
#ifdef CHATTY
					printf("processoutgoing(): Packet full adding OOB messages.\n");
#endif
					resume = 1;
					goto dispatch;
				} else {
					// pack the data in
					memcpy(workbufptr, peers[h].oobptr[ peers[h].oobqueuehead ],
							peers[h].ooblen[ peers[h].oobqueuehead ]);
					workbufptr += peers[h].ooblen[ peers[h].oobqueuehead ];

					msgcount++;
#ifdef CHATTY
					printf("processoutgoing(): Added %d oobbytes to packet\n",peers[h].ooblen[ peers[h].oobqueuehead ]);
#endif
					peers[h].oobqueuehead = (peers[h].oobqueuehead + 1) % OOBQUEUESZ;
				}
			}
			resume = 0;
			todo &= ~2;
		}

		if (todo & 4) {
			// calculate the space we need
			if (peers[h].acksequence > peers[h].sequence)
				k = NUM_SEQUENCES + peers[h].sequence - peers[h].acksequence;
			else
				k = peers[h].sequence - peers[h].acksequence;
			if (k > 0) k = 1+((k-1+7)>>3);

			i = j = -1;
			for (l=0; l<NUM_SEQUENCES; l++)
				if (peers[h].msgflags[l] & MSGFLAG_ACK) {
					if (i == -1)
						i = j = l;
					else
						j = l;
				}
			if (i == -1) l = 0;
			else l = 1+(((j-i+1)-1+7)>>3);

			if (k>0 || l>0) {
				if ((workbufptr - workbuf) + (k+l+1) > MAX_PACKET_LEN) {
					// lists won't fit, so send the current packet on its way
#ifdef CHATTY
					printf("processoutgoing(): Packet full adding ack lists.\n");
#endif
					resume = 1;
					goto dispatch;
				}
			
				// if any messages are unacknowledged add an unacknowledgement list
				*(workbufptr++) = k;

				i=0;
				for (s=peers[h].acksequence; s!=peers[h].sequence; s++)
					if (peers[h].outlen[s] > 0) {
						if (!i) {
							*(workbufptr++) = s;
#ifdef CHATTY
							printf("processoutgoing(): Added unacks: %02x", s);
#endif
							i++;
						} else {
							*workbufptr |= (1<<((i-1)&7));
#ifdef CHATTY
							printf(" %02x", i);
#endif
							i++;
							if (((i-1) & 7) == 0) workbufptr++;
						}
					}
				if (i > 0 && ((i-1) & 7) != 0) workbufptr++;
#ifdef CHATTY
				if (i>0) printf("\n");
#endif

				// if any messages need acknowledging add an acknowledgement list
				i=0;
				for (k=0; k<NUM_SEQUENCES; k++)
					if (peers[h].msgflags[k] & MSGFLAG_ACK) {
						if (!i) {
							*(workbufptr++) = k&255;
#ifdef CHATTY
							printf("processoutgoing(): Added acks: %02x", k&255);
#endif
							i++;
						} else {
							*workbufptr |= (1<<((i-1)&7));
#ifdef CHATTY
							printf(" %02x", i);
#endif
							i++;
							if (((i-1) & 7) == 0) workbufptr++;
						}
						peers[h].msgflags[k] &= MSGFLAG_ACK;
					}
				if (i > 0 && ((i-1) & 7) != 0) workbufptr++;
#ifdef CHATTY
				if (i>0) printf("\n");
#endif
			}

			resume = 0;
			todo &= ~4;
		}

dispatch:
		// if any data was added to the packet, checksum and send the packet on its way,
		// however, we only send it if some messages were added, or, the clock moved,
		// or we explicitly want the packet to be transmitted (ie. if we got a msg this frame)
		// otherwise we succeed only in flooding the network every frame with extra packets
		if ((workbufptr - workbuf) > PACKETHEADLEN &&
		    ((msgcount>0) || (myclock != lastclock) || (peers[h].flags & PEERFLAG_FORCETX))) {
			workbuf[4] = msgcount;
			
			crc = crc32(workbuf + 4, workbufptr - workbuf - 4);
			workbuf[0] = crc & 255;
			workbuf[1] = (crc>>8) & 255;
			workbuf[2] = (crc>>16) & 255;
			workbuf[3] = (crc>>24) & 255;

			networksend(&peers[h].address, workbuf, workbufptr - workbuf);

			if (resume) goto resumepoint;

			peers[h].flags &= ~PEERFLAG_FORCETX;
		}
	}
}


/*
 * Game-land interface
 */

//
// initmultiplayers() -- Start up multiplayer
//
void initmultiplayers(char damultioption, char dacomrateoption, char dapriority)
{
	int i,j;
#ifdef ENABLECOMMDUMPING
	char *dump;
#endif

	if (damultioption == 0 && multioption > 0)
		uninitmultiplayers();	// going back to singleplayer
	
	numplayers = 1; myconnectindex = 0;
	if (damultioption == 0) {
		connecthead = 0; connectpoint2[0] = -1;
	} else {
		connecthead = -1;
	}

	outmsghead = 0;
	inmsghead = 0;
	oobmsgqueuehead = oobmsgqueuetail = 0;
	fetchhead = 0;
	for (i=0;i<MAXMULTIPLAYERS;i++)
		resetpeer(i);

	multioption = damultioption;
	if (damultioption == 0) return;
	
	if (startnetwork() < 0) {
		initprintf("Network system initialization failed!\n");
		multioption = 0;	// failed initialising network so fall back to singleplayer
	}

	initcrc32table();

	lastclock = totalclock;

#ifdef ENABLECOMMDUMPING
	if ((dump = Bgetenv("BUILD_COMMDUMP")) != NULL) {
		dumpfh = Bfopen(dump, "w");
	}

	if (dumpfh)
		putdumpblock(0);
#endif
}

//
// uninitmultiplayers() -- Shut down multiplayer
//
void uninitmultiplayers(void)
{
	stopnetwork();

	multidumpstats();
	
	numplayers = 1; myconnectindex = 0;
	connecthead = 0; connectpoint2[0] = -1;
	multioption = 0;

#ifdef ENABLECOMMDUMPING
	if (dumpfh) {
		putdumpblock(4);
		Bfclose(dumpfh);
		dumpfh = NULL;
	}
#endif
}

//
// setpackettimeout() -- Unused
//
void setpackettimeout(long datimeoutcount, long daresendagaincount)
{
}

//
// sendlogon() -- Unused
//
void sendlogon(void)
{
}

//
// sendlogoff() -- Unused
//
void sendlogoff(void)
{
	long i;
	char tempbuf[2];

	tempbuf[0] = 255;
	tempbuf[1] = myconnectindex;
	for(i=connecthead;i>=0;i=connectpoint2[i])
		if (i != myconnectindex)
			sendpacket(i,tempbuf,2L);
}

//
// getoutputcirclesize() -- Unused
//
long getoutputcirclesize(void)
{
	return 0;
}

//
// setsocket() -- Unused
//
void setsocket(short newsocket)
{
}

//
// flushpackets() -- Unused
//
void flushpackets(void)
{
}

//
// sendpacket() -- The game interface for sending a message
//
void sendpacket(long other, char *bufptr, long messleng)
{
	unsigned char *p;
	int all;
#ifdef PEDANTIC
	int h;
#endif

	if (multioption == 0) return;
	if (messleng > MAX_MESSAGE_LEN || messleng <= 0 || !bufptr) return;

	if (other < 0) {
#ifdef CHATTY
		printf("sendpacket(): Broadcasting...\n");
#endif
		all = 1;
		other = connecthead;
	} else {
#ifdef PEDANTIC
		for (h=connecthead; h!=-1 && h!=other; h=connectpoint2[h]) ;
		if (h==-1) return;
#endif
		all = 0;
	}

	for (; other != -1; other = connectpoint2[other]) {
		if (all && other == myconnectindex) continue;
		if ((peers[other].sequence + 1) % NUM_SEQUENCES == peers[other].acksequence) {
#ifdef CHATTY
			printf("sendpacket(): Queue full for peer %d.\n", other);
#endif
			if (!all) break;
			continue;	// full!
		}

		p = allocatestorage(ALLOC_OUT, MESSAGEHEADLEN + messleng);
		if (!p) break;

		p[0] = peers[other].sequence;	// sequence
		p[1] = (messleng - 1) & 255;	// message length
		p[2] = FLAG_TYPE_SYNC;		// flag as synchronous

		memcpy(p+MESSAGEHEADLEN, bufptr, messleng);

		peers[other].outptr[ peers[other].sequence ] = p;
		peers[other].outlen[ peers[other].sequence ] = MESSAGEHEADLEN + messleng;
		peers[other].outage[ peers[other].sequence ] = 0;
		peers[other].msgflags[ peers[other].sequence ] |= MSGFLAG_TX;	// flag for transmission

#ifdef CHATTY
		printf("sendpacket(): Queued %d bytes to peer %d\n", messleng, other);
#endif

		//peers[other].sequence = (peers[other].sequence + 1) % NUM_SEQUENCES;
		peers[other].sequence++;	// because I'm using an unsigned char I'm exploiting the wraparound

		if (!all) break;
	}
}

//
// sendoobpacket() -- Sends an OOB message to a peer. This puts the
//   message in the outgoing queue of the peer.
//
void sendoobpacket(long other, char *bufptr, long messleng)
{
	unsigned char *p;
	int all;
	
	if (multioption == 0) return;
	if (messleng > MAX_MESSAGE_LEN || messleng <= 0 || !bufptr) return;
	
	if (other < 0) {
#ifdef CHATTY
		printf("sendoobpacket(): Broadcasting...\n");
#endif
		all = 1;
		other = connecthead;
	} else {
		all = 0;
	}
	
	for (; other != -1; other = connectpoint2[other]) {
		if (all && other == myconnectindex) continue;
		if ((peers[ other ].oobqueuetail + 1) % OOBQUEUESZ == peers[ other ].oobqueuehead) {
#ifdef CHATTY
			printf("sendoobpacket(): Queue full for peer %d.\n", other);
#endif
			if (!all) break;
			continue;	// full
		}

		p = allocatestorage(ALLOC_OUT, MESSAGEHEADLEN + messleng);
		if (!p) break;

		p[0] = 0;			// sequence
		p[1] = (messleng - 1) & 255;	// message length
		p[2] = FLAG_TYPE_OOB;		// flag as OOB

		memcpy(p+MESSAGEHEADLEN, bufptr, messleng);
	
		peers[ other ].oobptr[ peers[ other ].oobqueuetail ] = p;
		peers[ other ].ooblen[ peers[ other ].oobqueuetail ] = MESSAGEHEADLEN + messleng;
		peers[ other ].oobqueuetail = (peers[ other ].oobqueuetail + 1) % OOBQUEUESZ;

#ifdef CHATTY
		printf("sendoobpacket(): Queued %d OOB bytes to peer %d\n", messleng, other);
#endif

		if (!all) break;
	}
}

//
// sendoobpacketto() -- Sends an OOB message to an arbitrary address.
//   This immediately constructs a packet and sends it.
//
void sendoobpacketto(multiaddr *other, char *bufptr, long messleng)
{
	unsigned char workbuf[ PACKETHEADLEN + MESSAGEHEADLEN + MAX_MESSAGE_LEN ];
	unsigned char *workptr;
	long i;
#ifdef CHATTY
	char temp[64];
#endif

	if (multioption == 0) return;
	if (messleng > MAX_MESSAGE_LEN || messleng <= 0 || !other || !bufptr) return;
	
	workptr = workbuf + 4;	// leave out the CRC for a moment
	
	*(workptr++) = 1;	// message count

	*(workptr++) = 0;	// sequence
	*(workptr++) = (messleng - 1) & 255;	// message length
	*(workptr++) = FLAG_TYPE_OOB;	// flag as OOB

	memcpy(workptr, bufptr, messleng);
	workptr += messleng;

	// finally, the CRC
	i = crc32(workbuf + 4, workptr - workbuf - 4);
	workbuf[0] = i & 255;
	workbuf[1] = (i>>8) & 255;
	workbuf[2] = (i>>16) & 255;
	workbuf[3] = (i>>24) & 255;

#ifdef CHATTY
	multiaddrtostring(other, temp, 64);
	printf("sendoobpacketto(): Sending %d bytes to %s\n",temp);
#endif

	networksend(other, workbuf, workptr - workbuf);
}

//
// getpacket() -- The game interface for getting a message
//
short getpacket(long *other, char *bufptr)
{
	short h, l=0;
	
	if (multioption == 0) return 0;

	// process the incoming network packets
	processincoming();

	// copy out a packet into the user buffer and remove it from the queue
	h = fetchhead;
	do {
#ifdef CHATTY
		printf("getpacket(): Peer %d, we expect msg %d, need ack for msg %d\n", h, peers[h].insequence, peers[h].acksequence);
		printf("getpacket(): msg %d length is %d\n", peers[h].insequence, peers[h].inlen[ peers[h].insequence ]);
#endif
		if (peers[h].inlen[ peers[h].insequence ] > 0) {
			*other = h;
			memcpy(bufptr, peers[h].inptr[ peers[h].insequence ],
					peers[h].inlen[ peers[h].insequence ]);
			l = peers[h].inlen[ peers[h].insequence ];
			peers[h].inlen[ peers[h].insequence ] = 0;

#ifdef CHATTY
			printf("getpacket(): Dequeued %d byte message (%d).\n", l, peers[h].insequence);
#endif
	
			//peers[h].insequence = (peers[h].insequence + 1) % NUM_SEQUENCES;
			peers[h].insequence++;
		}

		// move to the next peer for the next time we fetch a message
		h = connectpoint2[h];
		if (h == -1) h = connecthead;
	} while (h != fetchhead && l == 0);

	fetchhead = h;

	// age messages and send any waiting packets
	processoutgoing();
	
	return l;
}

//
// fetchoobmessage() -- Pops the next OOB message from the queue
//
int fetchoobmessage(multiaddr *addr, char *bufptr, long *messleng)
{
	int rv=0;
#ifdef CHATTY
	char temp[64];
#endif

	if (multioption == 0) return 0;

	// process the incoming network packets
	processincoming();
	
	if (!addr || !bufptr || !messleng) {
		rv = -2;
		goto finish;
	}
	if (oobmsgqueuehead == oobmsgqueuetail) {
		rv = 0;
		goto finish;
	}

	if (*messleng < oobmsgqueue[oobmsgqueuehead].len) rv = -1;
	else {
		memcpy(addr, &oobmsgqueue[oobmsgqueuehead].addr, sizeof(multiaddr));
		memcpy(bufptr, oobmsgqueue[oobmsgqueuehead].ptr, oobmsgqueue[oobmsgqueuehead].len);
		*messleng = oobmsgqueue[oobmsgqueuehead].len;
		rv = 1;
	}

	oobmsgqueuehead = (oobmsgqueuehead + 1) % OOBQUEUESZ;

#ifdef CHATTY
	multiaddrtostring(addr, temp, 64);
	printf("fetchoobmessage(): Dequeued %d byte message from %s.\n",*messleng,temp);
#endif

finish:
	// age messages and send any waiting packets
	processoutgoing();

	return rv;
}

//
// genericmultifunction() -- Interface for sending multiplayer requests
//   using the COMMIT interface
//
#define  CMD_SEND               1
#define  CMD_GET                2
#define  CMD_SENDTOALL          3
#define  CMD_SENDTOALLOTHERS    4
#define  CMD_SCORE              5
void genericmultifunction(long other, char *bufptr, long messleng, long command)
{
	short p;
	
	switch (command) {
		case CMD_GET:	// kinda unused
			break;
				
		case CMD_SEND:
			if (other == 0) other = myconnectindex;
			else other--;
			sendpacket(other, bufptr, messleng);
			break;
			
		case CMD_SENDTOALL:
			for (p = connecthead; p != -1; p = connectpoint2[p])
				sendpacket(p, bufptr, messleng);
			break;
			
		case CMD_SENDTOALLOTHERS:
			for (p = connecthead; p != -1; p = connectpoint2[p])
				if (p != myconnectindex) sendpacket(p, bufptr, messleng);
			break;
			
		case CMD_SCORE:
			break;
	}
}

//
// whichpeer() -- Determines which peer, if any, an address belongs to
//
int whichpeer(multiaddr *a)
{
	int peer;

	if (multioption == 0) return -1;

	for (peer=connecthead; peer!=-1; peer=connectpoint2[peer])
		if (!multicompareaddr(&peers[peer].address, a))
			return peer;

	return -1;
}


/*
 * Internal network transport stuff
 */

//
// startnetwork() -- Initialises networking stuff.
//
static int startnetwork(void)
{
#ifdef WINDOWS
	WSADATA wsa;
	int rv;
	char *e;

	if (!winsockinited) {
		rv = WSAStartup(MAKEWORD(1,1), &wsa);
		if (rv) {
			switch (rv) {
				case WSASYSNOTREADY: e = "Networking subsystem not ready (WSASYSNOTREADY)"; break;
				case WSAVERNOTSUPPORTED: e = "Requested Winsock version not available (WSAVERNOTSUPPORTED)"; break;
				case WSAEINPROGRESS: e = "Operation in progress (WSAINPROGRESS)"; break;
				case WSAEPROCLIM: e = "Task limit reached (WSAPROCLIM)"; break;
				case WSAEFAULT: e = "Invalid data pointer passed (WSAEFAULT)"; break;
				default: e = "Unknown Winsock error"; break;
			}
			initprintf("Windows Sockets startup error: %s\n", e);
			return -1;
		}
	
		if (wsa.wVersion != MAKEWORD(1,1)) {
			initprintf("Windows Sockets startup error: Incorrect Winsock version returned (%d.%d)\n",
					LOBYTE(wsa.wVersion), HIBYTE(wsa.wVersion));
			WSACleanup();
			return -1;
		}

		initprintf("Windows Sockets initialised\n");
	        initprintf("  Available version:     %d.%d\n", LOBYTE(wsa.wVersion), HIBYTE(wsa.wVersion));
		initprintf("  Max supported version: %d.%d\n", LOBYTE(wsa.wHighVersion), HIBYTE(wsa.wHighVersion));
		initprintf("  Driver description:    %s\n", wsa.szDescription);
		initprintf("  System status:         %s\n", wsa.szSystemStatus);
		initprintf("  Max sockets:           %d\n", wsa.iMaxSockets);
		initprintf("  Max UDP datagram size: %d\n", wsa.iMaxUdpDg);
		winsockinited = 1;
	}
#endif

	if (setupmultigame() < 0) return -1;
	
	return 0;
}

//
// stopnetwork() -- Shuts down networking stuff.
//
static void stopnetwork(void)
{
	endmultigame();

#ifdef WINDOWS
	if (winsockinited) {
		WSACleanup();
		winsockinited = 0;
	}
#endif
}

//
// setupmultigame() -- Opens the socket and configures for playing
//
static int setupmultigame(void)
{
	int rv;
	struct sockaddr_in sadr;

	if (ipsocket == -1) {
		ipsocket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
		if (ipsocket < 0) return -1;
	}

	// enable non-blocking IO
	rv = 1;
#ifdef WINDOWS
	if (ioctlsocket(ipsocket, FIONBIO, (u_long*)&rv) < 0) {
#else
	if (ioctl(ipsocket, FIONBIO, &rv) < 0) {
#endif
		endmultigame();
		return -1;
	}
/*
	// enable broadcast ability
	rv = 1;
	if (setsockopt(ipsocket, SOL_SOCKET, SO_BROADCAST, (char *)&rv, sizeof(rv)) < 0) {
		endmultigame();
		return -1;
	}
*/	
	sadr.sin_family = AF_INET;
	sadr.sin_addr.s_addr = htonl(INADDR_ANY);
	sadr.sin_port = htons(ipport);
	if (bind(ipsocket, (struct sockaddr *)&sadr, sizeof(sadr)) < 0) {
		endmultigame();
		return -1;
	}

	return 0;
}

//
// endmultigame() -- Closes the socket
//
static void endmultigame(void)
{
	if (ipsocket != -1) {
		shutdown(ipsocket, SHUT_RDWR);
#ifdef WINDOWS
		closesocket(ipsocket);
#else
		close(ipsocket);
#endif
		ipsocket = -1;
	}
}

//
// networksend() -- Sends a buffer to an address.
//
static void networksend(multiaddr *addr, void *buf, unsigned long len)
{
#ifdef DUMPEACHPACKET
	char straddr[64];
	unsigned i;
#endif
	int err;

	if (!addr || !buf) return;

#ifdef DUMPEACHPACKET
	multiaddrtostring(addr, straddr, 64);
	printf("To:     %s\n"
	       "Length: %d\n"
	       "Contents:\n",
	       straddr, len);
	for (i=0;i<len;i++) {
		if (i % 8 == 0) printf("   %4x: ", i);
		printf("%02x ", ((unsigned char *)buf)[i]);
		if (i % 8 == 7) printf("\n");
	}
	if (len&7) printf("\n");
	printf("\n");
#endif
	
	if (addr->type == ADDR_NONE) return;
	if (addr->type == ADDR_IPV4) {
#ifndef FAKESEND
		if ((err = sendto(ipsocket, buf, len, 0, (struct sockaddr *)&addr->addr.ipv4, sizeof(struct sockaddr_in)))< 0) {
#ifdef CHATTY
#ifdef WINDOWS
			err = WSAGetLastError();
#else
			err = errno;
#endif
			printf("sendto() error %d\n", err);
#endif
			stats.txerrors++;
			return;
		}
#ifdef CHATTY
		printf("sendto() returned %d\n",err);
#endif
#else
		printf("Pretending to send %d bytes over IPv4.\n", len);
#endif
		stats.txbytes += err;
		stats.txpackets++;

#ifdef ENABLECOMMDUMPING
		if (dumpfh) {
			putdumpblock(2);
			Bfputc(sizeof(multiaddr), dumpfh);
			Bfwrite(addr, sizeof(multiaddr), 1, dumpfh);
			Bfputc(len & 255, dumpfh); Bfputc((len >> 8) & 255, dumpfh);
			Bfwrite(addr, len & 65535, 1, dumpfh);
		}
#endif
		return;
	}
}


//
// networkrecv() -- Receives network data into a buffer.
//
static int networkrecv(multiaddr *addr, void *buf, unsigned long *len)
{
#ifdef DUMPEACHPACKET
	char straddr[64];
	unsigned i;
#endif
	int fromlen;
	struct sockaddr *sad;
	int rv;
	
	if (!addr) return -1;

	sad = (struct sockaddr *)&addr->addr;

	fromlen = sizeof(addr->addr);
	rv = recvfrom(ipsocket, buf, *len, 0, sad, &fromlen);

	if (sad->sa_family == AF_INET)
		addr->type = ADDR_IPV4;
	else
		addr->type = ADDR_NONE;

	if (rv > 0) {
		*len = rv;
		stats.rxbytes += rv;
		stats.rxpackets++;

#ifdef DUMPEACHPACKET
		multiaddrtostring(addr, straddr, 64);
		printf("From:   %s\n"
		       "Length: %d\n"
		       "Contents:\n",
		       straddr, rv);
		for (i=0;i<rv;i++) {
			if (i % 8 == 0) printf("   %4x: ", i);
			printf("%02x ", ((unsigned char *)buf)[i]);
			if (i % 8 == 7) printf("\n");
		}
		if (rv&7) printf("\n");
		printf("\n");
#endif
	
		
#ifdef ENABLECOMMDUMPING
		if (dumpfh) {
			putdumpblock(3);
			Bfputc(sizeof(multiaddr), dumpfh);
			Bfwrite(addr, sizeof(multiaddr), 1, dumpfh);
			Bfputc(rv & 255, dumpfh); Bfputc((rv >> 8) & 255, dumpfh);
			Bfwrite(addr, rv & 65535, 1, dumpfh);
		}
#endif
	} else {
		*len = 0;
	}
	
	return rv;
}



/*
 * Game-land extras
 */

//
// multiaddrtostring() -- Converts the multiaddr address passed into a string
//   which is more human-readable
//
int multiaddrtostring(multiaddr *addr, char *buf, int len)
{
	if (!addr) return -1;
	
	if (addr->type == ADDR_IPV4) {
#ifdef WINDOWS
//		if (!WSAAddressToString((LPSOCKADDR)&addr->addr.ipv4, sizeof(struct sockaddr_in), NULL, buf, (LPDWORD)&len))
//			return 0;
		char *t;
		t = inet_ntoa(addr->addr.ipv4.sin_addr);
		if (t) {
			Bsnprintf(buf,len,"%s:%d",t,ntohs(addr->addr.ipv4.sin_port));
			return 0;
		}
#else
		char t[64];
		int l=64;
	
		if (inet_ntop(AF_INET, &addr->addr.ipv4, t, l)) {
			Bsnprintf(buf,len,"%s:%d",t,ntohs(addr->addr.ipv4.sin_port));
			return 0;
		}
#endif
	} else {
		Bstrncpy(buf,"Invalid",len);
	}

	return -1;
}


//
// multistringtoaddr() -- Resolves a hostname and stores the result in a multiaddr
//
static int multistringtoaddr(char *str, multiaddr *addr)
{
	struct hostent *he;
	int port = ipport, t;
	char *host, *p, *p2;

	host = Balloca(strlen(str)+1);
	strcpy(host,str);

	p = strchr(host,':');
	if (p) {
		*(p++) = 0;
		t = strtol(p, &p2, 10);
		if (p != p2 && *p2 == 0) port = t;
	}

	memset(addr, 0, sizeof(multiaddr));
	he = gethostbyname(host);
	if (!he) {
		switch (h_errno) {
			case HOST_NOT_FOUND: OSD_Printf("Name resolution error: %s: Host not found\n", host); break;
			case NO_DATA: OSD_Printf("Name resolution error: %s: No address for this hostname\n", host); break;
			case NO_RECOVERY: OSD_Printf("Name resolution error: %s: Unrecoverable error\n", host); break;
			case TRY_AGAIN: OSD_Printf("Name resolution error: %s: Temporary name server error\n", host); break;
		}
		
		return -1;
	}

	if (he->h_addrtype == AF_INET) {
		addr->type = ADDR_IPV4;
		addr->addr.ipv4.sin_family = AF_INET;
		memcpy(&addr->addr.ipv4.sin_addr, he->h_addr_list[0], he->h_length);
		addr->addr.ipv4.sin_port = htons(port);
	}

	return 0;
}


//
// multicompareaddr() -- Compares two addresses. Returns nonzero on difference.
//
int multicompareaddr(multiaddr *addr1, multiaddr *addr2)
{
	if (addr1->type != addr2->type) return 1;
	switch (addr1->type) {
		case ADDR_NONE:
			return 0;
		case ADDR_IPV4:
			if (addr1->addr.ipv4.sin_addr.s_addr != addr2->addr.ipv4.sin_addr.s_addr)
				return 1;
			if (addr1->addr.ipv4.sin_port != addr2->addr.ipv4.sin_port)
				return 1;
			return 0;
		default:
			return 1;
	}
}


//
// multiaddplayer() -- Hack trick to force the engine to recognize a player point
//
int multiaddplayer(char *str)
{
	short p;
	multiaddr addr;
	char temp[64];

	if (multistringtoaddr(str, &addr)) {
		initprintf("Failed adding %s to game.\n", str);
		return -1;
	}

	multiaddrtostring(&addr,temp,64);

	if (connecthead == -1) {
		connecthead = 0;
		connectpoint2[0] = -1;
		p = 0;
	} else {
		for (p = connecthead; connectpoint2[p] != -1; p = connectpoint2[p]) ;
		connectpoint2[p] = p+1;
		connectpoint2[p+1] = -1;
		numplayers++;
		p++;
	}

	memcpy(&peers[p].address, &addr, sizeof(multiaddr));

	initprintf("Added %s (%s) to game as player %d.\n", temp, str, numplayers-1);
	
#ifdef ENABLECOMMDUMPING
	if (dumpfh) {
		putdumpblock(1);
		Bfputc(Bstrlen(temp), dumpfh);
		Bfwrite(temp, Bstrlen(temp), 1, dumpfh);
	}
#endif

	return 0;
}


//
// multicollectplayers() -- Collects players for the game in the way Commit did under DOS
//    Returns 0 if all systems go, or 1 if the waiting was cancelled.
//
int multicollectplayers(void)
{
	short head, c;
	short ready[MAXMULTIPLAYERS];
	char workchar;
	long l;
	multiaddr addr;
	long olastclock;

	return 0;

	memset(ready, 0, sizeof(short)*MAXMULTIPLAYERS);
	ready[myconnectindex] = 1;

	initprintf("Waiting for %d more players. Press Escape to abort.\n",numplayers-1);
	olastclock = totalclock = 0;
	
	while (1) {
		if (handleevents()) {
			if (quitevent) {
				keystatus[1] = 1;
				quitevent = 0;
			}
		}

		if (keystatus[1]) {
			keystatus[1] = 0;
			initprintf("Aborting.\n");
			return 1;
		}

		if (totalclock - olastclock > 60) {
			olastclock = totalclock;

			workchar = 0x81;
#ifdef CHATTY
			printf("Ping!\n");
#endif
			sendoobpacket(-1, &workchar, 1);
		}

		l = 1;
		if (fetchoobmessage(&addr, &workchar, &l)) {
			head = whichpeer(&addr);
			if (head != -1 && workchar == 0x81) {
				if (!ready[head]) initprintf("Got acknowledgement from player %d.\n", head);
				ready[head] = 1;
				
				c = 0;
				for (head = connecthead; head != -1; head = connectpoint2[head])
					if (ready[head]) c++;
				if (c == numplayers) {
					initprintf("All players ready.\n");
					return 0;
				}
			}
		}
	}
}

//
// multidumpstats() -- Prints some networking code statistics to stdout
//
void multidumpstats(void)
{
	int i;
	char temp[64];
	
	printf("Networking statistics summary\n");

	if (multioption != 0) {
		for (i=connecthead; i!=-1; i=connectpoint2[i]) {
			multiaddrtostring(&peers[i].address, temp, 64);
			printf("Player %d: %s\n", i, temp);
		}
	}
	
	printf("TX Bytes/Packets/Errors:  %d/%d/%d\nRX Bytes/Packets/Errors:  %d/%d/%d\n",
			stats.txbytes,stats.txpackets,stats.txerrors,
			stats.rxbytes,stats.rxpackets,stats.rxerrors);
}


//
// multiidle() -- Massages the queues
//
void multiidle(void)
{
	//if (multioption == 0) return;
	//processoutgoing();
}

